"""
Demonstrates how to memory map a file into a bitarray.
"""
import os
import mmap

from bitarray import bitarray


filename = 'big.data'
filesize = 10_000_000

# create a large file with zeros
with open(filename, 'wb') as fo:
    fo.write(filesize * b'\0')

# open the file in binary read-write mode for mapping into a bitarray
with open(filename, 'r+b') as f:
    mapping = mmap.mmap(f.fileno(), 0)
    a = bitarray(buffer=mapping, endian='little')

    assert len(a) == 8 * filesize
    assert not a.any()  # no bits 1
    a[-1] = 1           # set the last bit in the array to 1

# open in binary read-only mode
with open(filename, 'rb') as fi:
    m = mmap.mmap(fi.fileno(), 0, access=mmap.ACCESS_READ)
    b = bitarray(buffer=m, endian='little')

    assert len(b) == 8 * filesize
    assert b.count() == 1  # only one bit is set
    assert b[-1] == 1      # the last one
    try:
        b[0] = 1           # TypeError: cannot modify read-only memory
    except TypeError:
        pass
    assert b[0] == 0       # wasn't changed, still 0


os.unlink(filename)
print("OK")



# issue 6
# http://www-graphics.stanford.edu/~seander/bithacks.html#NextBitPermutation

from bitarray import bitarray, get_default_endian
from bitarray.util import ba2int, int2ba, zeros

from math import comb


def all_perm(n, k, endian=None):
    """all_perm(n, k, endian=None) -> iterator

Return an iterator over all bitarrays of length `n` with `k` bits set to 1
in lexicographical order.
"""
    n = int(n)
    if n < 0:
        raise ValueError("length must be >= 0")
    k = int(k)
    if k < 0 or k > n:
        raise ValueError("number of set bits must be in range(0, n + 1)")

    if k == 0:
        yield zeros(n, endian)
        return

    v = (1 << k) - 1
    for _ in range(comb(n, k)):
        yield int2ba(v, length=n,
                endian=get_default_endian() if endian is None else endian)
        t = (v | (v - 1)) + 1
        v = t | ((((t & -t) // (v & -v)) >> 1) - 1)


def next_perm(a):
    """next_perm(bitarray) -> bitarray

Return the next lexicographical permutation.  The length and the number
of 1 bits in the bitarray is unchanged.  The integer value (`ba2int`) of the
next permutation will always increase, except when the cycle is completed (in
which case the lowest lexicographical permutation will be returned).
"""
    v = ba2int(a)
    if v == 0:
        return a
    t = (v | (v - 1)) + 1
    w = t | ((((t & -t) // (v & -v)) >> 1) - 1)
    try:
        return int2ba(w, length=len(a), endian=a.endian())
    except OverflowError:
        return a[::-1]

# ---------------------------------------------------------------------------

import unittest

from bitarray import frozenbitarray
from bitarray.util import urandom
from bitarray.test_bitarray import Util


class PermTests(unittest.TestCase, Util):

    def test_explicit_1(self):
        a = bitarray('00010011', 'big')
        for s in ['00010101', '00010110', '00011001',
                  '00011010', '00011100', '00100011']:
            a = next_perm(a)
            self.assertEqual(a.count(), 3)
            self.assertEqual(a, bitarray(s, 'big'))

    def test_explicit_2(self):
        for seq in (['0'], ['1'], ['00'], ['11'], ['01', '10'],
                    ['001', '010', '100'], ['011', '101', '110'],
                    ['0011', '0101', '0110', '1001', '1010', '1100']):
            a = bitarray(seq[0], 'big')
            for i in range(20):
                self.assertEqual(a, bitarray(seq[i % len(seq)]))
                a = next_perm(a)

    def test_all_same(self):
        for endian in 'little', 'big':
            for n in range(1, 30):
                for v in 0, 1:
                    a = bitarray(n, endian)
                    a.setall(v)
                    self.assertEqual(next_perm(a), a)

    def test_turnover(self):
        for a in [bitarray('11111110000', 'big'),
                  bitarray('0000001111111', 'little')]:
            self.assertEqual(next_perm(a), a[::-1])

    def test_large(self):
        a = bitarray('10010101010100100110010101110100111100101111', 'big')
        b = next_perm(a)
        c = bitarray('10010101010100100110010101110100111100110111')
        self.assertEqual(b, c)

    def test_errors(self):
        self.assertRaises(ValueError, next_perm, bitarray())
        self.assertRaises(TypeError, next_perm, '1')

    def check_all_perm(self, s):
        s1 = s.count(1)
        n = 0
        a = bitarray(s)
        coll = set()
        while 1:
            a = next_perm(a)
            coll.add(frozenbitarray(a))
            self.assertEqual(len(a), len(s))
            self.assertEqual(a.count(), s1)
            self.assertEqual(a.endian(), s.endian())
            n += 1
            if a == s:
                break
        self.assertEqual(n, comb(len(s), s1))
        self.assertEqual(len(coll), n)

    def check_order(self, a):
        i = -1
        for _ in range(comb(len(a), a.count())):
            i, j = ba2int(a), i
            self.assertTrue(i > j)
            a = next_perm(a)

    def test_few(self):
        for s in '0', '1', '00', '01', '111', '0011', '01011', '000000011':
            for endian in 'little', 'big':
                a = bitarray(s, endian)
                self.check_all_perm(a)
                a.sort(a.endian() == 'little')
                self.check_order(a)

    def test_random(self):
        for n in range(1, 10):
            a = urandom(n, self.random_endian())
            self.check_all_perm(a)
            a.sort(a.endian() == 'little')
            self.check_order(a)

    def test_all_perm_explicit(self):
        for n, k, res in [
                (0, 0, ['']),
                (1, 0, ['0']),
                (1, 1, ['1']),
                (2, 0, ['00']),
                (2, 1, ['01', '10']),
                (3, 2, ['011', '101', '110']),
                ]:
            self.assertEqual(list(all_perm(n, k, 'big')),
                             [bitarray(s) for s in res])

    def test_all_perm_1(self):
        n, k = 10, 5
        c = 0
        s = set()
        for a in all_perm(n, k, 'little'):
            self.assertIsType(a, 'bitarray')
            self.assertEqual(len(a), n)
            self.assertEqual(a.count(), k)
            s.add(frozenbitarray(a))
            c += 1
        self.assertEqual(c, comb(n, k))
        self.assertEqual(len(s), comb(n, k))

# ---------------------------------------------------------------------------

if __name__ == '__main__':
    unittest.main()



from bitarray import bitarray, bits2bytes


class PBM: # Portable Bitmap
    def __init__(self, w = 0, h = 0):
        self.size = w, h
        self.update()
        self.data = bitarray(self.bits, endian='big')

    def update(self):
        w, h = self.size
        self.bytes_per_row = bits2bytes(w)
        self.bits_per_row = 8 * self.bytes_per_row
        self.bytes = self.bytes_per_row * h
        self.bits = 8 * self.bytes

    def info(self):
        print('size: %d x %d' % self.size)
        print('bytes per row: %d' % self.bytes_per_row)
        print('bits per row: %d' % self.bits_per_row)
        print('bitarray: ' + repr(self.data.buffer_info()))

    def clear(self):
        self.data.setall(0)

    def save(self, filename):
        with open(filename, 'wb') as fo:
            fo.write(b'P4\n')
            fo.write(b'# This is a partable bitmap (pbm) file.\n')
            fo.write(('%d %d\n' % self.size).encode())
            self.data.tofile(fo)

    def load(self, filename):
        with open(filename, 'rb') as fi:
            assert fi.readline().strip() == b'P4'
            while True:
                line = fi.readline()
                if not line.startswith(b'#'):
                    self.size = tuple(map(int, line.split()))
                    break
            self.update()
            self.data = bitarray(endian='big')
            self.data.fromfile(fi)
            fi.close()
        assert self.data.nbytes == self.bytes

    def address(self, pt):
        x, y = pt
        return x + self.bits_per_row * y

    def __getitem__(self, pt):
        return self.data[self.address(pt)]

    def __setitem__(self, pt, val):
        self.data[self.address(pt)] = val


if __name__ == '__main__':
    # draw picture with straight line from (10, 10) to (390, 390)
    a = PBM(500, 400)
    a.info()
    a.clear()
    for x in range(10, 391):
        a[x, x] = True
    a.save('pic1.ppm')

    # copy the picture
    b = PBM()
    b.load('pic1.ppm')
    b.save('pic2.ppm')

    # draw a straight line from (490, 10) to (110, 390) on top
    for i in range(381):
        b[490 - i, 10 + i] = 1
    b.save('pic3.ppm')



from time import time

from bitarray import bitarray, get_default_endian
from bitarray.util import urandom, ba2hex, hex2ba, ba2base, base2ba


# ----- conversion using prefix codes

CODEDICT = {'little': {}, 'big': {
    '0': bitarray('0000'),    '1': bitarray('0001'),
    '2': bitarray('0010'),    '3': bitarray('0011'),
    '4': bitarray('0100'),    '5': bitarray('0101'),
    '6': bitarray('0110'),    '7': bitarray('0111'),
    '8': bitarray('1000'),    '9': bitarray('1001'),
    'a': bitarray('1010'),    'b': bitarray('1011'),
    'c': bitarray('1100'),    'd': bitarray('1101'),
    'e': bitarray('1110'),    'f': bitarray('1111'),
}}
for k, v in CODEDICT['big'].items(): # type: ignore
    CODEDICT['little'][k] = v[::-1]  # type: ignore

def prefix_ba2hex(a):
    return ''.join(a.iterdecode(CODEDICT[a.endian()]))

def prefix_hex2ba(s, endian=None):
    a = bitarray(0, endian or get_default_endian())
    a.encode(CODEDICT[a.endian()], s)
    return a

# ----- conversion using ba2base (element based utility function)

def ba2_base16(a):
    return ba2base(16, a)

def base16_2ba(s, endian=None):
    return base2ba(16, s, endian)

# ----- test

def test_round(f, g, n, endian):
    # f: function which takes bitarray and returns hexstr
    # g: function which takes hexstr and returns bitarray
    # n: size of random bitarray
    a = urandom(n, endian)
    t0 = time()
    s = f(a)
    print('%s:  %9.6f sec' % (f.__name__, time() - t0))
    t0 = time()
    b = g(s, endian)
    print('%s:  %9.6f sec' % (g.__name__, time() - t0))
    assert b == a

if __name__ == '__main__':
    n = 100 * 1000 * 1000 + 4
    for endian in 'little', 'big':
        print('%s-endian:' % endian)
        for f in ba2hex, ba2_base16, prefix_ba2hex:
            for g in hex2ba, base16_2ba, prefix_hex2ba:
                test_round(f, g, n, endian)
            print()



"""
This program demonstrates how Huffman codes can be used to efficiently
compress and uncompress files (text or binary).
"""
import os
import struct
from itertools import islice
from optparse import OptionParser
from collections import Counter

from bitarray import bitarray
from bitarray.util import (serialize, deserialize,
                           canonical_huffman, canonical_decode)

def create_code(cnt):
    if len(cnt) > 0:
        return canonical_huffman(cnt)
    # special case for empty file
    return {0: bitarray('0')}, [0, 1], [0]

def encode(filename):
    with open(filename, 'rb') as fi:
        plain = fi.read()

    code, count, symbol = create_code(Counter(plain))
    with open(filename + '.huff2', 'wb') as fo:
        fo.write(struct.pack("<B", len(count)))
        for i in range(1, len(count)):
            fo.write(struct.pack("<H", count[i]))
        fo.write(struct.pack("<H", len(symbol)))
        fo.write(bytearray(symbol))

        a = bitarray(endian='little')
        a.encode(code, plain)
        fo.write(serialize(a))

    if len(plain) > 0:
        print('Bits: %d / %d' % (len(a), 8 * len(plain)))
        print('Ratio =%6.2f%%' % (100.0 * a.buffer_info()[1] / len(plain)))

def decode(filename):
    assert filename.endswith('.huff2')

    with open(filename, 'rb') as fi:
        stream = iter(fi.read())

    maxbits = struct.unpack("<B", bytes(islice(stream, 1)))[0]
    count = [0] + [struct.unpack("<H", bytes(islice(stream, 2)))[0]
                   for _ in range(maxbits - 1)]

    symbol_n = struct.unpack("<H", bytes(islice(stream, 2)))[0]
    symbol = bytearray(islice(stream, symbol_n))

    a = deserialize(bytes(stream))
    with open(filename[:-6] + '.out', 'wb') as fo:
        fo.write(bytes(canonical_decode(a, count, symbol)))

def main():
    p = OptionParser("usage: %prog [options] FILE")
    p.add_option(
        '-e', '--encode',
        action="store_true",
        help="encode (compress) FILE using the Huffman code calculated for "
             "the frequency of characters in FILE itself. "
             "The output is FILE.huff2 which contains both the Huffman "
             "code and the bitarray resulting from the encoding.")
    p.add_option(
        '-d', '--decode',
        action="store_true",
        help="decode (decompress) FILE.huff2 and write the output to FILE.out")
    p.add_option(
        '-t', '--test',
        action="store_true",
        help="encode FILE, decode FILE.huff2, compare FILE with FILE.out, "
             "and unlink created files.")
    opts, args = p.parse_args()
    if len(args) != 1:
        p.error('exactly one argument required')
    filename = args[0]

    if opts.encode:
        encode(filename)

    elif opts.decode:
        decode(filename + '.huff2')

    elif opts.test:
        huff = filename + '.huff2'
        out = filename + '.out'
        encode(filename)
        decode(huff)
        assert open(filename, 'rb').read() == open(out, 'rb').read()
        os.unlink(huff)
        os.unlink(out)

    else:
        p.error("no option provided")


if __name__ == '__main__':
    main()



import sys

if sys.version_info[:2] < (3, 6):
    sys.exit("This example requires Python 3.6 or higher")

try:
    from setuptools import setup, Extension
except ImportError:
    from distutils.core import setup, Extension


setup(
    name = "puff",
    ext_modules = [Extension(
        name = "_puff",
        sources = ["_puff.c"],
        include_dirs = ["../../bitarray"],
    )],
)



import json
from base64 import b64encode, b64decode

from bitarray import bitarray
from bitarray.util import serialize, deserialize


class JSONEncoder(json.JSONEncoder):

    def default(self, obj):

        if isinstance(obj, bitarray):
            if len(obj) > 50:
                return {'bitarray_b64': b64encode(serialize(obj)).decode()}
            else:
                return {'bitarray': obj.to01()}

        return json.JSONEncoder.default(self, obj)


class JSONDecoder(json.JSONDecoder):

    def __init__(self, *args, **kwargs):
        json.JSONDecoder.__init__(self, object_hook=self.object_hook,
                                  *args, **kwargs)

    def object_hook(self, obj):
        if isinstance(obj, dict) and len(obj) == 1:
            if 'bitarray_b64' in obj:
                return deserialize(b64decode(obj['bitarray_b64']))

            if 'bitarray' in obj:
                return bitarray(obj['bitarray'])

        return obj


def test():
    from random import randint
    from bitarray.util import urandom

    a = [urandom(n * n, endian=['little', 'big'][randint(0, 1)])
         for n in range(12)]
    a.append({'key1': bitarray('010'),
              'key2': 'value2',
              'key3': urandom(300)})
    j = JSONEncoder(indent=2).encode(a)
    print(j)

    b = JSONDecoder().decode(j)
    assert a == b
    assert b[-1]['key1'] == bitarray('010')


if __name__ == '__main__':
    test()



"""
Implementation of a sparse bitarray

Internally we store a list of positions at which a bit changes from
1 to 0 or vice versa.  Moreover, we start with bit 0, meaning that if the
first bit in the bitarray is 1 our list starts with posistion 0.
For example:

   bitarray('110011111000')

is represented as:

   flips:   [0, 2, 4, 9, 12]

The last element in the list is always the length of the bitarray, such that
an empty bitarray is represented as [0].
"""
from bisect import bisect, bisect_left

from bitarray import bitarray

from common import Common


class SparseBitarray(Common):

    def __init__(self, x = 0):
        if isinstance(x, int):
            self.flips = [x]  # bitarray with x zeros
        else:
            self.flips = [0]
            for v in x:
                self.append(int(v))

    def __len__(self):
        return self.flips[-1]

    def __getitem__(self, key):
        if isinstance(key, slice):
            start, stop = self._get_start_stop(key)
            if stop <= start:
                return SparseBitarray()

            i = bisect(self.flips, start)
            j = bisect_left(self.flips, stop)

            res = SparseBitarray()
            res.flips = [0] if i % 2 else []
            for k in range(i, j):
                res.flips.append(self.flips[k] - start)
            res.flips.append(stop - start)
            return res

        elif isinstance(key, int):
            if not 0 <= key < len(self):
                raise IndexError
            return bisect(self.flips, key) % 2

        else:
            raise TypeError

    def __setitem__(self, key, value):
        if isinstance(key, slice):
            start, stop = self._get_start_stop(key)
            if stop <= start:
                return

            i = bisect(self.flips, start)
            j = bisect_left(self.flips, stop)

            self.flips[i:j] = (
                ([] if i % 2 == value else [start]) +
                ([] if j % 2 == value else [stop])
            )

        elif isinstance(key, int):
            if not 0 <= key < len(self):
                raise IndexError
            p = bisect(self.flips, key)
            if p % 2 == value:
                return
            self.flips[p:p] = [key, key + 1]

        else:
            raise TypeError

        self._reduce()

    def __delitem__(self, key):
        if isinstance(key, slice):
            start, stop = self._get_start_stop(key)
            if stop <= start:
                return

            i = bisect(self.flips, start)
            j = bisect_left(self.flips, stop)

            size = stop - start
            for k in range(j, len(self.flips)):
                self.flips[k] -= size
            self.flips[i:j] = [start] if (j - i) % 2 else []

        elif isinstance(key, int):
            if not 0 <= key < len(self):
                raise IndexError
            p = bisect(self.flips, key)
            for j in range(p, len(self.flips)):
                self.flips[j] -= 1

        else:
            raise TypeError

        self._reduce()

    def _reduce(self):
        n = self.flips[-1]      # length of bitarray
        lst = []                # new representation list
        i = 0
        while True:
            c = self.flips[i]   # current element (at index i)
            if c == n:          # element with bitarray length reached
                break
            j = i + 1           # find next value (at index j)
            while self.flips[j] == c:
                j += 1
            if (j - i) % 2:     # only append index if repeated odd times
                lst.append(c)
            i = j
        lst.append(n)
        self.flips = lst

    def _intervals(self):
        v = 0
        start = 0
        for stop in self.flips:
            yield v, start, stop
            v = 1 - v
            start = stop

    def append(self, value):
        if value == len(self.flips) % 2:  # opposite value as last element
            self.flips.append(len(self) + 1)
        else:                             # same value as last element
            self.flips[-1] += 1

    def extend(self, other):
        n = len(self)
        m = len(other.flips)
        if len(self.flips) % 2:
            self.flips.append(n)
        for i in range(m):
            self.flips.append(other.flips[i] + n)
        self._reduce()

    def find(self, value):
        if len(self) == 0:
            return -1
        flips = self.flips
        if value:
            return -1 if len(flips) == 1 else flips[0]
        else:
            if flips[0] > 0:
                return 0
            return -1 if len(flips) == 2 else flips[1]

    def to_bitarray(self):
        a = bitarray(len(self))
        for v, start, stop in self._intervals():
            a[start:stop] = v
        return a

    def invert(self):
        self.flips.insert(0, 0)
        self._reduce()

    def insert(self, i, value):
        i = self._adjust_index(i)
        p = bisect_left(self.flips, i)
        for j in range(p, len(self.flips)):
            self.flips[j] += 1
        self[i] = value

    def count(self, value=1):
        cnt = 0
        for v, start, stop in self._intervals():
            if v == value:
                cnt += stop - start
        return cnt

    def reverse(self):
        n = len(self)
        lst = [0] if len(self.flips) % 2 else []
        lst.extend(n - p for p in reversed(self.flips))
        lst.append(n)
        self.flips = lst
        self._reduce()



"""
This library contains useful functionality for working with Huffman trees
and codes.

Note:
There is a function for directly creating a Huffman code from a frequency
map in the bitarray library itself: bitarray.util.huffman_code()
"""
from bitarray import bitarray


class Node(object):
    def __init__(self):
        self.child = [None, None]
        self.freq = None

    def __lt__(self, other):
        # heapq needs to be able to compare the nodes
        return self.freq < other.freq


def huff_code(tree):
    """
    Given a Huffman tree, traverse the tree and return the Huffman code, i.e.
    a dictionary mapping symbols to bitarrays.
    """
    result = {}

    def traverse(nd, prefix=bitarray()):
        try:  # leaf
            result[nd.symbol] = prefix
        except AttributeError:
            traverse(nd.child[0], prefix + bitarray([0]))
            traverse(nd.child[1], prefix + bitarray([1]))

    traverse(tree)
    return result


def insert_symbol(tree, ba, sym):
    """
    Insert symbol into a tree at the position described by the bitarray,
    creating nodes as necessary.
    """
    nd = tree
    for k in ba:
        prev = nd
        nd = nd.child[k]

        if hasattr(nd, 'symbol'):
            raise ValueError("ambiguity")

        if nd is None:
            nd = Node()
            prev.child[k] = nd

    if hasattr(nd, 'symbol') or nd.child[0] or nd.child[1]:
        raise ValueError("ambiguity")

    nd.symbol = sym


def make_tree(codedict):
    """
    Create a tree from the given code dictionary, and return its root node.
    Unlike trees created by huff_tree, all nodes will have .freq set to None.
    """
    tree = Node()
    for sym, ba in codedict.items():
        insert_symbol(tree, ba, sym)
    return tree


def traverse(tree, it):
    """
    Traverse tree until a leaf node is reached, and return its symbol.
    This function consumes an iterator on which next() is called during each
    step of traversing.
    """
    nd = tree
    while 1:
        nd = nd.child[next(it)]
        if nd is None:
            raise ValueError("prefix code does not match data in bitarray")

        try:
            return nd.symbol
        except AttributeError:
            pass

    if nd != tree:
        raise ValueError("decoding not terminated")


def iterdecode(tree, bitsequence):
    """
    Given a tree and a bitsequence, decode the bitsequence and generate
    the symbols.
    """
    it = iter(bitsequence)
    while True:
        try:
            yield traverse(tree, it)
        except StopIteration:
            return


def write_dot(tree, fn, binary=False):
    """
    Given a tree (which may or may not contain frequencies), write
    a graphviz '.dot' file with a visual representation of the tree.
    """
    special_ascii = {' ': 'SPACE', '\n': 'LF', '\r': 'CR', '\t': 'TAB',
                     '\\': r'\\', '"': r'\"'}
    def disp_sym(i):
        if binary:
            return '0x%02x' % i
        else:
            c = chr(i)
            res = special_ascii.get(c, c)
            assert res.strip(), repr(c)
            return res

    def disp_freq(f):
        if f is None:
            return ''
        return '%d' % f

    def write_nd(fo, nd):
        if hasattr(nd, 'symbol'):  # leaf node
            a, b = disp_freq(nd.freq), disp_sym(nd.symbol)
            fo.write('  %d  [label="%s%s%s"];\n' %
                     (id(nd), a, ': ' if a and b else '', b))
            return

        assert hasattr(nd, 'child')
        fo.write('  %d  [shape=circle, style=filled, '
                 'fillcolor=grey, label="%s"];\n' %
                 (id(nd), disp_freq(nd.freq)))

        for k in range(2):
            if nd.child[k]:
                fo.write('  %d->%d;\n' % (id(nd), id(nd.child[k])))

        for k in range(2):
            if nd.child[k]:
                write_nd(fo, nd.child[k])

    with open(fn, 'w') as fo:    # dot -Tpng tree.dot -O
        fo.write('digraph BT {\n')
        fo.write('  node [shape=box, fontsize=20, fontname="Arial"];\n')
        write_nd(fo, tree)
        fo.write('}\n')


def print_code(freq, codedict):
    """
    Given a frequency map (dictionary mapping symbols to their frequency)
    and a codedict, print them in a readable form.
    """
    special_ascii = {0: 'NUL', 9: 'TAB', 10: 'LF', 13: 'CR', 127: 'DEL'}
    def disp_char(i):
        if 32 <= i < 127:
            return repr(chr(i))
        return special_ascii.get(i, '')

    print(' symbol     char    hex   frequency     Huffman code')
    print(70 * '-')
    for i in sorted(codedict, key=lambda c: (freq[c], c), reverse=True):
        print('%7r     %-4s    0x%02x %10i     %s' % (
            i, disp_char(i), i, freq[i], codedict[i].to01()))


def test():
    from bitarray.util import _huffman_tree

    freq = {'a': 10, 'b': 2, 'c': 1}
    tree = _huffman_tree(freq)
    code = huff_code(tree)
    assert len(code['a']) == 1
    assert len(code['b']) == len(code['c']) == 2

    code = {'a': bitarray('0'),
            'b': bitarray('10'),
            'c': bitarray('11')}
    tree = make_tree(code)
    txt = 'abca'
    a = bitarray()
    a.encode(code, txt)
    assert a == bitarray('010110')
    assert ''.join(iterdecode(tree, a)) == txt


if __name__ == '__main__':
    test()



from bitarray import bitarray


def show(a):
    info = a.buffer_info()
    size = info[1]
    alloc = info[4]
    print('%d  %d' % (size, alloc))

a = bitarray()
prev = -1
while len(a) < 2000:
    alloc = a.buffer_info()[4]
    if prev != alloc:
        show(a)
    prev = alloc
    a.append(1)

for i in 800_000, 400_000, 399_992, 0, 0, 80_000:
    if len(a) < i:
        a.extend(bitarray(i - len(a)))
    else:
        del a[i:]
    assert len(a) == i
    show(a)



import sys
assert sys.version_info[0] == 3, "This program requires Python 3"

import re
from doctest import testfile
from io import StringIO

import bitarray.util


BASE_URL = "https://github.com/ilanschnell/bitarray"

NEW_IN = {
    'bitarray':               '2.3: optional `buffer` argument',
    'bitarray.bytereverse':   '2.2.5: optional start and stop arguments',
    'bitarray.clear':         '1.4',
    'bitarray.count':        ['1.1.0: optional start and stop arguments',
                              '2.3.7: optional step argument'],
    'bitarray.find':          '2.1',
    'bitarray.frombytes':     '2.5.0: allow bytes-like argument',
    'bitarray.invert':        '1.5.3: optional index argument',
    'bitarray.pack':          '2.5.0: allow bytes-like argument',
    'decodetree':             '1.6',
    'frozenbitarray':         '1.1',
    'get_default_endian':     '1.3',
    'util.any_and':           '2.7',
    'util.ba2base':           '1.9',
    'util.base2ba':           '1.9',
    'util.count_n':           '2.3.6: optional value argument',
    'util.deserialize':      ['1.8',
                              '2.5.0: allow bytes-like argument'],
    'util.intervals':         '2.7',
    'util.make_endian':       '1.3',
    'util.parity':            '1.9',
    'util.pprint':            '1.8',
    'util.rindex':            '2.3.0: optional start and stop arguments',
    'util.serialize':         '1.8',
    'util.urandom':           '1.7',
    'util.sc_encode':         '2.7',
    'util.sc_decode':         '2.7',
    'util.vl_decode':         '2.2',
    'util.vl_encode':         '2.2',
    'util.canonical_huffman': '2.5',
    'util.canonical_decode':  '2.5',
}

DOCS = {
    'chc': ('Canonical Huffman Coding', 'canonical.rst'),
    'rep': ('Bitarray representations', 'represent.rst'),
    'sc':  ('Compression of sparse bitarrays', 'sparse_compression.rst'),
    'vlf': ('Variable length bitarray format', 'variable_length.rst'),
}

DOC_LINKS = {
    'util.canonical_huffman':  'chc',
    'util.canonical_decode':   'chc',
    'util.ba2base':            'rep',
    'util.base2ba':            'rep',
    'util.deserialize':        'rep',
    'util.serialize':          'rep',
    'util.sc_decode':          'sc',
    'util.sc_encode':          'sc',
    'util.vl_decode':          'vlf',
    'util.vl_encode':          'vlf',
}

NOTES = {
    'bitarray.pack': """\
   This method, as well as the ``.unpack()`` method, are meant for efficient
   transfer of data between bitarray objects to other Python objects (for
   example NumPy's ndarray object) which have a different memory view.""",

    'bitarray.tolist': """\
   Note that the list object being created will require 32 or 64 times more
   memory (depending on the machine architecture) than the bitarray object,
   which may cause a memory error if the bitarray is very large.""",

    'util.count_xor': "   This is also known as the Hamming distance."
}

GETSET = {
    'bitarray.nbytes':     'int',
    'bitarray.padbits':    'int',
    'bitarray.readonly':   'bool',
}

_NAMES = set()

sig_pat = re.compile(r"""
(                # group 1
  (\w+)          # function name, group 2
  \([^()]*\)     # (...)
)
(                # optional group 3
  \s->\s(.+)     # return type, group 4
)?
""", re.VERBOSE)

def get_doc(name):
    parts = name.split('.')
    obj = bitarray
    while parts:
        obj = getattr(obj, parts.pop(0))

    lines = obj.__doc__.splitlines()

    if len(lines) == 1:
        sig = '``%s`` -> %s' % (obj.__name__, GETSET[name])
        return sig, lines

    m = sig_pat.match(lines[0])
    if m is None:
        raise Exception("signature invalid: %r" % lines[0])
    sig = '``%s``' %  m.group(1)
    assert m.group(2) == obj.__name__
    if m.group(4):
        sig += ' -> %s' % m.group(4)
    assert lines[1] == ''
    return sig, lines[2:]


def write_doc(fo, name):
    _NAMES.add(name)
    sig, lines = get_doc(name)
    fo.write(sig + '\n')
    for line in lines:
        out = line.rstrip()
        fo.write("   %s\n" % out.replace('`', '``') if out else "\n")

    note = NOTES.get(name)
    if note:
        fo.write("\n%s\n" % note)

    link = DOC_LINKS.get(name)
    if link:
        title, filename = DOCS[link]
        url = BASE_URL + '/blob/master/doc/' + filename
        fo.write("\n   See also: `%s <%s>`__\n" % (title, url))

    new_in = NEW_IN.get(name)
    if new_in:
        for line in new_in if isinstance(new_in, list) else [new_in]:
            fo.write("\n   New in version %s.\n" % line.replace('`', '``'))

    fo.write('\n\n')


def write_reference(fo):
    fo.write("""\
Reference
=========

bitarray version: %s -- `change log <%s>`__

In the following, ``item`` and ``value`` are usually a single bit -
an integer 0 or 1.


The bitarray object:
--------------------

""" % (bitarray.__version__, BASE_URL + "/blob/master/doc/changelog.rst"))
    write_doc(fo, 'bitarray')

    fo.write("bitarray methods:\n"
             "-----------------\n\n")
    for method in sorted(dir(bitarray.bitarray)):
        if method.startswith('_'):
            continue
        name = 'bitarray.%s' % method
        if name not in GETSET:
            write_doc(fo, name)

    fo.write("bitarray data descriptors:\n"
             "--------------------------\n\n"
             "Data descriptors were added in version 2.6.\n\n")
    for getset in sorted(dir(bitarray.bitarray)):
        name = 'bitarray.%s' % getset
        if name in GETSET:
            write_doc(fo, name)

    fo.write("Other objects:\n"
             "--------------\n\n")
    write_doc(fo, 'frozenbitarray')
    write_doc(fo, 'decodetree')

    fo.write("Functions defined in the `bitarray` module:\n"
             "-------------------------------------------\n\n")
    for func in sorted(['test', 'bits2bytes', 'get_default_endian']):
        write_doc(fo, func)

    fo.write("Functions defined in `bitarray.util` module:\n"
             "--------------------------------------------\n\n"
             "This sub-module was added in version 1.2.\n\n")
    for func in bitarray.util.__all__:
        write_doc(fo, 'util.%s' % func)

    for name in list(NEW_IN) + list(DOC_LINKS):
        assert name in _NAMES, name


def update_readme(path):
    ver_pat = re.compile(r'(bitarray.+?)\s(\d+\.\d+\.\d+)')

    with open(path, 'r') as fi:
        data = fi.read()

    with StringIO() as fo:
        for line in data.splitlines():
            if line == 'Reference':
                break
            line = ver_pat.sub(r'\1 ' + bitarray.__version__, line)
            fo.write("%s\n" % line.rstrip())

        write_reference(fo)
        new_data = fo.getvalue()

    if new_data == data:
        print("already up-to-date")
    else:
        with open(path, 'w') as f:
            f.write(new_data)


def write_changelog(fo):
    ver_pat = re.compile(r'(\d{4}-\d{2}-\d{2})\s+(\d+\.\d+\.\d+)')
    hash_pat = re.compile(r'#([0-9a-f]+)')
    link_pat = re.compile(r'\[(.+)\]\((.+)\)')

    def hash_replace(match):
        group1 = match.group(1)
        if len(group1) >= 7:
            if len(group1) != 8:
                print("Warning: commit hash length != 8, got", len(group1))
            url = "%s/commit/%s" % (BASE_URL, group1)
        else:
            url = "%s/issues/%d" % (BASE_URL, int(group1))
        return "`%s <%s>`__" % (match.group(0), url)

    fo.write("Change log\n"
             "==========\n\n")

    for line in open('./CHANGE_LOG'):
        line = line.rstrip()
        match = ver_pat.match(line)
        if match:
            line = match.expand(r'**\2** (\1):')
        elif line.startswith('-----'):
            line = ''
        elif line.startswith('  '):
            line = line[2:]
            line = line.replace('`', '``')
            line = hash_pat.sub(hash_replace, line)
            line = link_pat.sub(r"`\1 <\2>`__", line)
        fo.write(line + '\n')


def main():
    if len(sys.argv) > 1:
        sys.exit("no arguments expected")

    update_readme('./README.rst')
    with open('./doc/reference.rst', 'w') as fo:
        write_reference(fo)
    with open('./doc/changelog.rst', 'w') as fo:
        write_changelog(fo)

    testfile('./README.rst')
    testfile('./doc/buffer.rst')
    testfile('./doc/canonical.rst')
    testfile('./doc/represent.rst')
    testfile('./doc/sparse_compression.rst')
    testfile('./doc/variable_length.rst')


if __name__ == '__main__':
    main()



from __future__ import print_function
from bitarray.util import urandom, pprint, ba2base, base2ba


a = urandom(60)

for m in range(1, 7):
    n = 1 << m
    print("----- length: %d ----- base: %d ----- " % (m, n))
    pprint(a, group=m)
    rep = ba2base(n, a)
    print("representation:", rep)
    print()
    assert base2ba(n, rep) == a



import sys
from bitarray import bitarray
from numba import jit  # type: ignore

width, height = 4000, 3000
maxdepth = 500


@jit(nopython=True)
def mandel(c):
    d = 0
    z = c
    while abs(z) < 4.0 and d <= maxdepth:
        d += 1
        z = z * z + c
    return d


def main():
    data = bitarray(endian='big')

    for j in range(height):
        sys.stdout.write('.')
        sys.stdout.flush()
        y = +1.5 - 3.0 * j / height
        for i in range(width):
            x = -2.75 + 4.0 * i / width
            c = mandel(complex(x, y)) % 2
            data.append(c)
    print("done")

    with open('out.ppm', 'wb') as fo:
        fo.write(b'P4\n')
        fo.write(b'# partable bitmap image of the Mandelbrot set\n')
        fo.write(b'%i %i\n' % (width, height))
        data.tofile(fo)


if __name__ == '__main__':
    main()



# gene sequence example from @yoch, see
# https://github.com/ilanschnell/bitarray/pull/54

from random import choice
from timeit import timeit

from bitarray import bitarray


trans = {
    "A": bitarray("00"),
    "T": bitarray("01"),
    "G": bitarray("10"),
    "C": bitarray("11")
}

N = 10000
seq = [choice("ATGC") for _ in range(N)]

arr = bitarray()
arr.encode(trans, seq)

assert arr.decode(trans) == seq

# decodage
t = timeit(lambda: arr.decode(trans), number=1000)
print(t)



from bitarray import bitarray


def show(a):
    info = a.buffer_info()
    print('%18d %10d %10d' % (info[0], info[1], info[4]))


# make sure sequence of appends will always increase allocated size
a = bitarray()
prev = -1
while len(a) < 1_000_000:
    alloc = a.buffer_info()[4]
    assert prev <= alloc
    prev = alloc
    a.append(1)


# ensure that when we start from a large array and delete part, we always
# get a decreasing allocation
a = bitarray(10_000_000)
prev = a.buffer_info()[4]
for _ in range(100):
    del a[-100_000:]
    alloc = a.buffer_info()[4]
    assert alloc <= prev
    prev = alloc


# initalizing a bitarray from a list or bitarray should not overallocate
for n in range(1000):
    alloc = (n + 3) & ~3
    assert n <= alloc < n + 4 and alloc % 4 == 0
    a = bitarray(8 * n * [1])
    assert alloc == a.buffer_info()[4]
    b = bitarray(a)
    assert alloc == b.buffer_info()[4]


# starting from a large bitarray, make we sure we don't realloc each time
# we extend
a = bitarray(1_000_000)  # no overallocation
assert a.buffer_info()[4] == 125_000
a.extend(bitarray(8))  # overallocation happens here
alloc = a.buffer_info()[4]
for _ in range(1000):
    a.extend(bitarray(8))
    assert a.buffer_info()[4] == alloc


print("OK")



#
# This example illusatrates how binary data can be efficiently be passed
# between a bitarray object and an ndarray with dtype bool
#
from __future__ import print_function

import bitarray
import numpy  # type: ignore

a = bitarray.bitarray('100011001001')
print(a)

# bitarray  ->  ndarray
b = numpy.frombuffer(a.unpack(), dtype=bool)
print(repr(b))

# ndarray  ->  bitarray
c = bitarray.bitarray()
c.pack(b.tobytes())

assert a == c



from random import random, randint
from time import time

from bitarray import bitarray, decodetree
from bitarray.util import huffman_code


N = 1000 * 1000

# create Huffman code for N symbols
code = huffman_code({i: random() for i in range(N)})
print(len(code))

# create the decodetree object
t0 = time()
tree = decodetree(code)
print('decodetree(code):  %9.6f sec' % (time() - t0))

print(tree.nodes())
plain = [randint(0, N - 1) for _ in range(100)]

a = bitarray()
a.encode(code, plain)

# decode using the code dictionary
t0 = time()
res = a.decode(code)
print('decode(code):  %9.6f sec' % (time() - t0))
assert res == plain

# decode using the decodetree
t0 = time()
res = a.decode(tree)
print('decode(tree):  %9.6f sec' % (time() - t0))
assert res == plain
assert tree.todict() == code



"""
Demonstrates the implementation of "Sieve of Eratosthenes" algorithm for
finding all prime numbers up to any given limit.
"""
from math import ceil, sqrt

from bitarray import bitarray
from bitarray.util import count_n


MILLION = 1000 * 1000
N = 100 * MILLION

# Each bit a[i] corresponds to whether or not i is a prime
a = bitarray(N)
a.setall(True)
# Zero and one are not prime
a[:2] = False
# Perform sieve
for i in range(2, ceil(sqrt(N))):
    if a[i]:  # i is prime, so all multiples are not
        a[i*i::i] = False

print('the first few primes are:')
print(a.search(1, 20))

# There are 5,761,455 primes up to 100 million
print('there are %d primes up to %d' % (a.count(), N))

# The number of twin primes up to 100 million is 440,312
print('number of twin primes up to %d is %d' %
      (N, len(a.search(bitarray('101')))))

# The 1 millionth prime number is 15,485,863
m = MILLION
print('the %dth prime is %d' % (m, count_n(a, m) - 1))



from collections import Counter, defaultdict

from bitarray import bitarray

from _puff import State, MAXLCODES, MAXDCODES, FIXLCODES


# fixed literal/lengths and distance lengths
FIXED_LENGTHS = tuple(
    # literal/length lengths (FIXLCODES elements)
    [8] * 144 + [9] * 112 + [7] * 24 + [8] * 8 +
    # distance lengths (MAXDCODES elements)
    [5] * MAXDCODES
)

class Puff(State):

    def process_blocks(self, callback=None):
        self.stats = defaultdict(Counter)

        while True:
            # read the three bit block header
            last = self.read_uint(1)   # 1 if last block
            btype = self.read_uint(2)  # block type
            self.stats['btype'][btype] += 1

            if btype == 0:                      # process stored block
                self.process_stored_block()

            elif btype == 1:                    # process fixed block
                self.decode_block(FIXED_LENGTHS, FIXLCODES, MAXDCODES)

            elif btype == 2:                    # process dynamic block
                self.process_dynamic_block()

            else:
                raise ValueError("Reserved block type: %d" % btype)

            if callback:
                callback(self.stats)

            if last:
                break

        return self.stats

    def process_stored_block(self):  # uncompressed block
        self.align_byte_boundary()

        # read length
        blen: int = self.read_uint(16)
        nlen: int = self.read_uint(16)
        if blen ^ 0xffff != nlen:
            raise ValueError("Invalid length in uncompressed block")

        self.stats['stored blen'][blen] += 1

        # copy bytes
        self.extend_block(blen)

    def process_dynamic_block(self) -> None:
        # permutation of code length codes
        order = [16, 17, 18, 0, 8, 7, 9, 6, 10, 5,
                 11, 4, 12, 3, 13, 2, 14, 1, 15]

        # get number of lengths in each table, check lengths
        nlen = self.read_uint(5) + 257
        ndist = self.read_uint(5) + 1
        ncode = self.read_uint(4) + 4
        if nlen > MAXLCODES or ndist > MAXDCODES:
            raise ValueError("bad counts")

        self.stats['dynamic nlen'][nlen] += 1
        self.stats['dynamic ndist'][ndist] += 1
        self.stats['dynamic ncode'][ncode] += 1

        # read code length code lengths (really), missing lengths are zero
        lengths = 19 * [0]
        for index in range(ncode):
            lengths[order[index]] = self.read_uint(3)

        # decode literal/lengths and distance lengths
        lengths = self.decode_lengths(lengths, nlen + ndist)

        # decode actual block data
        self.decode_block(lengths, nlen, ndist)

    def align_byte_boundary(self):
        # discard bits to align to byte boundary
        skip = 8 - self.get_incnt() % 8  # bits to skip
        if skip != 8:
            self.read_uint(skip)



from bitarray import bitarray
from bitarray.util import int2ba, ba2int, pprint


class SmallIntArray(object):
    """
    A class which allows efficiently storeing an array of integers
    represented by a specified number of bits.
    For example, an array with 1000 5 bit integers can be created,
    allowing each element in the array to take values form 0 to 31,
    while the size of the object is 625 (5000/8) bytes.
    """
    def __init__(self, N, k):
        self.N = N  # number of integers
        self.k = k  # bits for each integer
        self.array = bitarray(N * k)

    def slice_i(self, i):
        assert 0 <= i < self.N
        return slice(self.k * i, self.k * (i + 1))

    def __getitem__(self, i):
        return ba2int(self.array[self.slice_i(i)])

    def __setitem__(self, i, v):
        self.array[self.slice_i(i)] = int2ba(v, self.k)


if __name__ == '__main__':
    from random import randint

    # define array with 1000 integers, each represented by 5 bits
    a = SmallIntArray(1000, 5)

    b = [] # store values, for assertion below
    for i in range(1000):
        v = randint(0, 31)
        b.append(v)
        a[i] = v

    print(b[:5])
    print(a.array.buffer_info())
    pprint(a.array[:25], group=5)

    for i in range(1000):
        assert a[i] == b[i]



"""
The purpose of this script is to illustrate how copy_n() in _bitarray.c works.
This is essentially a Python implementation of copy_n() with output of the
different stages of the bitarray we copy into.
For more details, see also: bitarray/copy_n.txt
"""
import sys
from random import randint
from io import StringIO

from bitarray import bitarray, bits2bytes
from bitarray.util import pprint, urandom


verbose = False

def mark_range_n(i, n, c, text=''):
    a = bitarray(i * '0' + n * '1')
    f = StringIO()
    pprint(a, stream=f)
    s = f.getvalue()
    print("%-10s" % text + ''.join(c if e == '1' else ' ' for e in s[10:]))


def mark_range(i, j, c, text=''):
    mark_range_n(i, j - i, c, text)


def shift_r8(self, a, b, n):
    assert a <= b and 0 <= n < 8
    self[8 * a : 8 * b] >>= n

def is_be(self):
    return self.endian() == 'big'

bitmask_table = [
    [0x01, 0x02, 0x04, 0x08, 0x10, 0x20, 0x40, 0x80],  # little endian
    [0x80, 0x40, 0x20, 0x10, 0x08, 0x04, 0x02, 0x01],  # big endian
]

ones_table = [
    [0x00, 0x01, 0x03, 0x07, 0x0f, 0x1f, 0x3f, 0x7f],  # little endian
    [0x00, 0x80, 0xc0, 0xe0, 0xf0, 0xf8, 0xfc, 0xfe],  # big endian
]

def copy_n(self, a, other, b, n):
    assert 0 <= a <= len(self)
    assert 0 <= b <= len(other)
    assert n >= 0
    if n == 0 or (self is other and a == b):
        return

    if a % 8 == 0 and b % 8 == 0:            # aligned case
        p1 = a // 8
        p2 = (a + n - 1) // 8
        m = bits2bytes(n)

        assert p1 + m == p2 + 1
        m2 = ones_table[is_be(self)][(a + n) % 8]
        t2 = memoryview(self)[p2]

        memoryview(self)[p1:p1 + m] = memoryview(other)[b // 8:b // 8 + m]
        if self.endian() != other.endian():
            self.bytereverse(p1, p2 + 1)

        if m2:
            memoryview(self)[p2] = (memoryview(self)[p2] & m2) | (t2 & ~m2)
        return

    if n < 8:                                # small n case
        if a <= b:  # loop forward
            for i in range(n):
                self[i + a] = other[i + b]
        else:       # loop backwards
            for i in range(n - 1, -1, -1):
                self[i + a] = other[i + b]
        return

    # -------------------------------------- # general case
    p1 = a // 8
    p2 = (a + n - 1) // 8
    p3 = b // 8
    sa = a % 8
    sb = -(b % 8)
    m1 = ones_table[is_be(self)][sa]
    m2 = ones_table[is_be(self)][(a + n) % 8]

    assert n >= 8
    assert a - sa == 8 * p1
    assert b + sb == 8 * p3
    assert a + n > 8 * p2

    if verbose:
        print('a =', a)
        print('b =', b)
        print('n =', n)
        print('p1 =', p1)
        print('p2 =', p2)
        print('p3 =', p3)
        print('sa =', sa)
        print('sb =', sb)

    t1 = memoryview(self)[p1]
    t2 = memoryview(self)[p2]
    t3 = memoryview(other)[p3]

    if sa + sb < 0:
        sb += 8
        if verbose:
            print(' -> sb =', sb)

    if verbose:
        print('other')
        pprint(other)
        mark_range_n(b, n, '^', 'b..b+n')
        mark_range_n(b + sb, n - sb, '=')
        mark_range_n(b, sb, '3')

        print('self')
        pprint(self)
        mark_range_n(a, n, '^', 'a..a+n')
        mark_range(8 * p1, a, '1')
        mark_range(a + n, 8 * p2 + 8, '2')

        print('copy_n')
        mark_range_n(a - sa, n - sb, '=')

    copy_n(self, a - sa, other, b + sb, n - sb)
    if verbose:
        pprint(self)

        print('rshift', sa + sb)
        mark_range(8 * p1, 8 * (p2 + 1), '>')

    shift_r8(self, p1, p2 + 1, sa + sb)
    if verbose:
        pprint(self)
        mark_range_n(8 * p1 + sa + sb, n - sb, '=', 'a..a+n')

    if m1:
        if verbose:
            mark_range(8 * p1, a, '1')
        memoryview(self)[p1] = (memoryview(self)[p1] & ~m1) | (t1 & m1)

    if m2 and sa + sb:
        if verbose:
            mark_range(a + n, 8 * p2 + 8, '2')
        memoryview(self)[p2] = (memoryview(self)[p2] & m2) | (t2 & ~m2)

    if verbose:
        mark_range_n(a, sb, '3')
    for i in range(sb):
        self[i + a] = bool(t3 & bitmask_table[is_be(other)][(i + b) % 8])

    if verbose:
        pprint(self)


def test_copy_n():

    def random_endian():
        return ['little', 'big'][randint(0, 1)]

    for N in range(1000):
        M = randint(0, 5 + 2 * N)
        n = randint(0, min(N, M))
        a = randint(0, N - n)
        b = randint(0, M - n)
        x = urandom(N, random_endian())
        y = urandom(M, random_endian())
        z = x.copy()
        copy_n(x, a, y, b, n)
        z[a:a + n] = y[b:b + n]
        assert x == z

    for N in range(1000):
        n = randint(0, N)
        a = randint(0, N - n)
        b = randint(0, N - n)
        x = urandom(N, random_endian())
        z = x.copy()
        copy_n(x, a, x, b, n)
        z[a:a + n] = z[b:b + n]
        assert x == z


if __name__ == '__main__':
    test_copy_n()
    verbose = True
    other = bitarray(
        '00101110 11111001 01011101 11001011 10110000 01011110 011')
    self =  bitarray(
        '01011101 11100101 01110101 01011001 01110100 10001010 01111011')
    copy_n(self, 21, other, 6, 31)
    assert self == bitarray(
        '01011101 11100101 01110101 11110010 10111011 10010111 01101011')



import os
import sys
import zlib
import unittest

from bitarray import bitarray

from puff import State, Puff, FIXLCODES, MAXDCODES, FIXED_LENGTHS


class TestState(unittest.TestCase):

    def test_simple(self):
        a = bitarray(80)
        b = bytearray()
        s = State(a, b)
        self.assertEqual(s.get_incnt(), 0)
        self.assertEqual(len(b), 0)
        s.extend_block(4)
        self.assertEqual(s.get_incnt(), 32)
        self.assertEqual(len(b), 4)
        a[32:35] = bitarray('011')
        self.assertEqual(s.read_uint(3), 6)
        self.assertEqual(s.get_incnt(), 35)

    def test_read_uint(self):
        # works for either bit endianness
        inp = bitarray('11011100 1')
        out = bytearray()
        s = State(inp, out)
        self.assertRaises(ValueError, s.read_uint, -1)  # negative bits
        self.assertEqual(s.read_uint(0), 0)  # reading zero bits is OK
        self.assertEqual(s.read_uint(3), 3)
        self.assertEqual(s.read_uint(5), 7)
        self.assertEqual(s.get_incnt(), 8)
        self.assertRaises(ValueError, s.read_uint, 2)  # end of input
        self.assertEqual(s.read_uint(1), 1)
        self.assertEqual(s.read_uint(0), 0)
        self.assertEqual(s.get_incnt(), 9)
        self.assertRaises(ValueError, s.read_uint, 1)  # end of input
        self.assertEqual(len(out), 0)                  # nothing in output

    def test_read_uint32(self):
        a = bitarray(endian='little')
        a.frombytes(b'\x7e\xae\xd4\xbb')
        s = State(a, bytearray())
        self.assertEqual(s.read_uint(32), 0xbbd4ae7e)
        self.assertEqual(s.get_incnt(), 32)

        a = bitarray(32 * '1', endian='little')
        s = State(a, bytearray())
        self.assertEqual(s.read_uint(32), (1 << 32) - 1)
        self.assertEqual(s.get_incnt(), 32)

    def test_copy(self):
        a = bitarray()  # nothing is read from input in this test
        out = bytearray(b'ABC')
        s = State(a, out)
        s.copy(3, 2)
        self.assertEqual(bytes(out), b'ABCAB')
        self.assertRaises(ValueError, s.copy, 6, 1)   # distance too far back
        s.copy(5, 10)
        s.copy(6, 0)  # does nothing (length is zero)
        self.assertEqual(bytes(out), b'ABCABABCABABCAB')
        self.assertRaises(ValueError, s.copy, 0, 1)   # distance zero
        self.assertRaises(ValueError, s.copy, -1, 1)  # distance negative
        self.assertRaises(ValueError, s.copy, 16, 1)  # distance too far back
        self.assertRaises(ValueError, s.copy, 1, -1)  # length negative

    def test_append_byte(self):
        out = bytearray()
        s = State(bitarray(), out)
        s.append_byte(0)
        self.assertRaises(ValueError, s.append_byte, -1)
        self.assertRaises(ValueError, s.append_byte, 256)
        s.append_byte(255)
        self.assertEqual(bytes(out), b'\0\xff')

    def test_extend_block(self):
        a = bitarray()
        a.frombytes(b'ABCDEF')
        b = bytearray()
        s = State(a, b)

        s.extend_block(0)
        self.assertEqual(bytes(b), b'')

        s.extend_block(2)
        self.assertEqual(bytes(b), b'AB')
        self.assertRaises(ValueError, s.extend_block, 5)  # not enough input

        s.extend_block(1)
        self.assertEqual(bytes(b), b'ABC')
        self.assertEqual(s.get_incnt(), 24)

        s.read_uint(3)
        self.assertRaises(ValueError, s.extend_block, 1)  # input unaligned
        s.read_uint(5)
        s.extend_block(2)
        self.assertEqual(bytes(b), b'ABCEF')  # we skipped 'D'
        # invalid block size
        self.assertRaises(ValueError, s.extend_block, -1)
        self.assertRaises(ValueError, s.extend_block, 0x10000)

    def test_decode_lengths(self):
        # this is taken from the stream of dynamic header bits - after nlen,
        # ndist, ncode and the (up to 19) code length code lengths are read
        a = bitarray('''
    11001100 00001100 00011101 11011101 11000011 00000111 00001000 00101100
    00011100 10000011 01100000 11101110 11101011 00000111 01011000 00111011
    10111100 00000010 00011000 00111011 10111010 01000010 00001110 11100001
    00000111 01110000 10010010 00001001 10000011 10110100 00100001 10000011
    10111011 10101100 00011101 11011100 00100111 10111111 11011001 11011110
    11111011 01010111 10110100 11111010 11101010 11101010 10101110 11110100
    01011110 10001110 01010101 11010101 01011111 11010111 1100
        ''')
        ncode = 279 + 23

        b = bytearray()
        s = State(a, b)
        length = s.decode_lengths([4, 0, 6, 5, 4, 0, 0, 4, 2, 3,
                                   6, 0, 5, 5, 0, 0, 2, 4, 0], ncode)
        # no bytes were added to the output stream
        self.assertEqual(len(b), 0)
        # the code lengths list contains literal/lengths and distance codes
        self.assertEqual(len(length), ncode)
        # we've exhausted the input array exactly
        self.assertEqual(s.get_incnt(), len(a))
        # simple sum check, as I didn't want to cut and paste the whole list
        self.assertEqual(sum(length), 2183)

    def test_decode_lengths_error(self):
        a = bitarray(1000)
        b = bytearray()
        s = State(a, b)
        lengths = 19 * [0]
        # nlen > 316 (MAXCODES)
        self.assertRaises(ValueError, s.decode_lengths, lengths, 317)
        # sequence length not 19
        self.assertRaises(ValueError, s.decode_lengths, 20 * [0], 316)
        lengths[1] = 16
        # length[1] > MAXBITS
        self.assertRaises(ValueError, s.decode_lengths, lengths, 316)

    def test_decode_block_error(self):
        a = bitarray(1000)
        b = bytearray()
        s = State(a, b)
        lengths = 302 * [0]
        # nlen > 288 (FIXLCODES)
        self.assertRaises(ValueError, s.decode_block, lengths, 289, 23)
        # ndist > 30 (MAXDCODES)
        self.assertRaises(ValueError, s.decode_block, lengths, 279, 31)
        # sequence length not 279 + 23 = 302
        self.assertRaises(ValueError, s.decode_block, 301 * [0], 279, 23)
        lengths[1] = 16
        # length[1] > MAXBITS
        self.assertRaises(ValueError, s.decode_block, lengths, 279, 23)


class TestFixedBlock(unittest.TestCase):

    @staticmethod
    def decode(a, buffer=b''):
        res = bytearray(buffer)
        s = State(a, res)
        s.decode_block(FIXED_LENGTHS, FIXLCODES, MAXDCODES)
        return bytes(res)

    def test_literal(self):
        a = bitarray('01111001 10011100 10010001 10011110 0000000')
        #             I        l        a        n        end-of-block
        self.assertEqual(self.decode(a), b"Ilan")

    def test_rle(self):
        a = bitarray('01110001 0000001  00000   0000000')
        #             A        len=3    dist=1  end-of-block
        self.assertEqual(self.decode(a), b"AAAA")

    def test_rle_258(self):
        a = bitarray('01110001 11000101  00000   0000000')
        #             A        len=258   dist=1  end-of-block
        self.assertEqual(self.decode(a), 259 * b"A")

        a = bitarray('01110010 11000100 11111  00000   0000000')
        #             B        len=227  31     dist=1  end-of-block
        # here len = 227 + 31 = 258, same as before
        self.assertEqual(self.decode(a), 259 * b"B")

    def test_max_back(self):
        a = bitarray('0000001  11101 1111111111111  0000000')
        #             len=3    dist=24577 + 8191
        buffer = b'ABCD' + 32764 * b'-'
        self.assertEqual(len(buffer), 1 << 15)
        out = self.decode(a, buffer)
        self.assertEqual(out, buffer + b'ABC')

    def test_too_far_back(self):
        a = bitarray('01110001 0000001  00001   0000000')
        #             A        len=3    dist=2  end-of-block
        self.assertRaises(ValueError, self.decode, a)

    def test_invalid_length_symbols(self):
        a = bitarray('11000110')  # symbol 286
        self.assertRaises(ValueError, self.decode, a)
        a = bitarray('11000111')  # symbol 287
        self.assertRaises(ValueError, self.decode, a)


class TestPuff(unittest.TestCase):

    def test_constants(self):
        self.assertEqual(len(FIXED_LENGTHS), FIXLCODES + MAXDCODES)

    def test_align_byte_boundary(self):
        a = bitarray(15)
        d = Puff(a, bytearray())
        d.read_uint(5)
        d.align_byte_boundary()
        self.assertEqual(d.get_incnt(), 8)
        d.align_byte_boundary()
        self.assertEqual(d.get_incnt(), 8)
        d.read_uint(1)
        self.assertRaises(ValueError, d.align_byte_boundary)

    def round_trip(self, data, level=-1):
        compressed = zlib.compress(data, level=level)

        a = bitarray(0, 'little')
        a.frombytes(compressed)
        out = bytearray()
        p = Puff(a, out)
        # check zlib header
        self.assertEqual(p.read_uint(8), 0x78)
        self.assertTrue(p.read_uint(8) in (0x01,   # no compression
                                           0x5e,   # low compression
                                           0x9c,   # default compression
                                           0xda))  # best compression
        p.process_blocks()

        self.assertEqual(bytes(out), data)

    def test_zeros(self):
        for n in 0, 1, 10, 100, 1000, 10_000:
            self.round_trip(n * b'\0')

    def test_this_file(self):
        with open(__file__, 'rb') as f:
            data = f.read()
        for level in range(1, 10):
            self.round_trip(data, level)

    @unittest.skipIf(sys.platform == "win32", "Windows doesn't have this file")
    def test_words(self):
        with open('/usr/share/dict/words', 'rb') as f:
            data = f.read()
        self.round_trip(data)

    def test_random_data(self):
        self.round_trip(os.urandom(2000))


if __name__ == '__main__':
    unittest.main()



from optparse import OptionParser
from time import time
from collections import Counter

from bitarray import bitarray
from bitarray.util import _huffman_tree

from huffman import (huff_code, write_dot, print_code,
                     make_tree, iterdecode)


def main():
    p = OptionParser("usage: %prog [options] [FILE]")
    p.add_option(
        '-p', '--print',
        action="store_true",
        help="print Huffman code")
    p.add_option(
        '-t', '--tree',
        action="store_true",
        help="store the tree as a .dot file")
    opts, args = p.parse_args()

    if len(args) == 0:
        filename = 'README'
    elif len(args) == 1:
        filename = args[0]
    else:
        p.error('only one argument expected')

    with open(filename, 'rb') as fi:
        plain = bytearray(fi.read())

    t0 = time()
    freq = Counter(plain)
    print('count:     %9.6f sec' % (time() - t0))

    t0 = time()
    tree = _huffman_tree(freq)
    print('tree:      %9.6f sec' % (time() - t0))

    if opts.tree:
        write_dot(tree, 'tree.dot', 0 in plain)
    code = huff_code(tree)
    if opts.print:
        print_code(freq, code)
    if opts.tree:
        # create tree from code (no frequencies)
        write_dot(make_tree(code), 'tree_raw.dot', 0 in plain)

    a = bitarray()

    t0 = time()
    a.encode(code, plain)
    print('C encode:  %9.6f sec' % (time() - t0))

    # Time the decode function above
    t0 = time()
    res = bytearray(iterdecode(tree, a))
    Py_time = time() - t0
    print('Py decode: %9.6f sec' % Py_time)
    assert res == plain

    # Time the decode method which is implemented in C
    t0 = time()
    res = bytearray(a.iterdecode(code))
    C_time = time() - t0
    print('C decode:  %9.6f sec' % C_time)
    assert res == plain

    print('Ratio: %f' % (Py_time / C_time))


if __name__ == '__main__':
    main()



import sys
from random import randint
import unittest

try:
    from itertools import pairwise  # type: ignore
except ImportError:
    from itertools import tee
    def pairwise(iterable):
        a, b = tee(iterable)
        next(b, None)
        return zip(a, b)

from bitarray import bitarray
from bitarray.util import intervals
from bitarray.test_bitarray import Util

if len(sys.argv) != 2 or sys.argv[1] not in ('flips', 'ones', '-'):
    sys.exit("Argument 'flips' or 'ones' expected.")
MODE = sys.argv[1]
del sys.argv[1]


class TestsSparse(unittest.TestCase, Util):

    def check(self, s, a):
        if MODE == 'flips':
            self.assertTrue(len(s.flips) > 0)
            self.assertTrue(s.flips[0] >= 0)
            for x, y in pairwise(s.flips):
                self.assertTrue(y > x)
            self.assertEqual(s.to_bitarray(), a)

        elif MODE == 'ones':
            if s.ones:
                self.assertTrue(s.ones[-1] < s.n)
            for x, y in pairwise(s.ones):
                self.assertTrue(y > x)
            self.assertEqual(s.to_bitarray(), a)

        else:
            self.assertEqual(s, a)

    def test_init(self):
        if MODE != '-':
            for n in 0, 1, 2, 3, 99:
                a = bitarray(n)
                a.setall(0)
                t = BitArray(n)
                self.check(t, a)

        for s in '', '0', '1', '01110001':
            a = bitarray(s)
            t = BitArray(s)
            self.check(t, a)

    def test_repr(self):
        s = BitArray('01001')
        if MODE != '-':
            self.assertEqual(repr(s), "SparseBitarray('01001')")

    def test_len(self):
        for a in self.randombitarrays():
            s = BitArray(a)
            self.assertEqual(len(s), len(a))

    def test_getitem_index(self):
        for a in self.randombitarrays(start=1):
            s = BitArray(a)
            for i in range(len(a)):
                self.assertEqual(s[i], a[i])

    def test_getitem_slice(self):
        for a in self.randombitarrays():
            s = BitArray(a)
            i = randint(0, len(s))
            j = randint(0, len(s))
            self.check(s[i:j], a[i:j])

    def test_setitem_index(self):
        for a in self.randombitarrays(start=1):
            s = BitArray(a)
            for _ in range(10):
                i = randint(0, len(s) - 1)
                v = randint(0, 1)
                s[i] = a[i] = v
                self.check(s, a)

    def test_setitem_slice(self):
        for a in self.randombitarrays():
            s = BitArray(a)
            for _ in range(10):
                i = randint(0, len(s))
                j = randint(0, len(s))
                v = randint(0, 1)
                s[i:j] = a[i:j] = v
                self.check(s, a)

    def test_delitem_index(self):
        for a in self.randombitarrays(start=1):
            s = BitArray(a)
            i = randint(0, len(s) - 1)
            del s[i]
            del a[i]
            self.check(s, a)

    def test_delitem_slice(self):
        for a in self.randombitarrays():
            s = BitArray(a)
            i = randint(0, len(s))
            j = randint(0, len(s))
            del s[i:j]
            del a[i:j]
            self.check(s, a)

    def test_append(self):
        for a in self.randombitarrays():
            s = BitArray()
            for v in a:
                s.append(v)
            self.check(s, a)

    def test_find(self):
        for a in self.randombitarrays():
            s = BitArray(a)
            for v in 0, 1:
                self.assertEqual(s.find(v), a.find(v))

    def test_extent(self):
        for aa in self.randombitarrays():
            for b in self.randombitarrays():
                a = aa.copy()
                s = BitArray(a)
                t = BitArray(b)
                s.extend(t)
                a.extend(b)
                self.check(s, a)

            s = BitArray(aa)
            s.extend(s)
            self.check(s, 2 * aa)

    def test_count(self):
        for a in self.randombitarrays():
            s = BitArray(a)
            for v in 0, 1:
                self.assertEqual(s.count(v), a.count(v))

    def test_insert(self):
        for a in self.randombitarrays():
            s = BitArray(a)
            i = randint(-2, len(s) + 2)
            v = randint(0, 1)
            s.insert(i, v)
            a.insert(i, v)
            self.check(s, a)

    def test_invert(self):
        for a in self.randombitarrays():
            s = BitArray(a)
            s.invert()
            a.invert()
            self.check(s, a)

    def test_pop(self):
        for a in self.randombitarrays(start=1):
            s = BitArray(a)
            i = randint(-len(a), len(a) - 1)
            self.assertEqual(s.pop(i), a.pop(i))
            self.check(s, a)

    def test_remove(self):
        for a in self.randombitarrays():
            s = BitArray(a)
            v = randint(0, 1)
            error = 0
            try:
                s.remove(v)
            except ValueError:
                error += 1
            try:
                a.remove(v)
            except ValueError:
                error += 1
            self.assertTrue(error % 2 == 0)
            self.check(s, a)

    def test_reverse(self):
        for a in self.randombitarrays():
            s = BitArray(a)
            s.reverse()
            a.reverse()
            self.check(s, a)

    def test_sort(self):
        for a in self.randombitarrays():
            s = BitArray(a)
            for rev in 0, 1:
                s.sort(rev)
                a.sort(rev)
                self.check(s, a)

    if MODE == 'flips':
        def test_flips(self):
            for a in self.randombitarrays():
                lst = [] if a and a[0] == 0 else [0]
                lst.extend(t[2] for t in intervals(a))
                s = BitArray(a)
                self.assertEqual(s.flips, lst)

        def test_reduce(self):
            for a, b in [
                    ([0],                 [0]),
                    ([0, 0],              [0]),
                    ([3, 7],              [3, 7]),
                    ([3, 7, 7],           [3, 7]),
                    ([3, 3, 7, 7, 7],     [7]),
                    ([3, 3, 3, 7, 7],     [3, 7]),
                    ([0, 0, 2, 2],        [2]),
                    ([0, 2, 2, 2, 2, 3],  [0, 3]),
                    ([0, 0, 0, 1, 1, 2, 2, 2, 3, 4, 4, 4, 4, 5],
                     [0, 2, 3, 5]),
                ]:
                s = BitArray()
                s.flips = a
                s._reduce()
                self.assertEqual(s.flips, b)


if __name__ == '__main__':
    if MODE == '-':
        BitArray = bitarray
    else:
        BitArray = __import__(MODE).SparseBitarray  # type: ignore
    unittest.main()



import sys
import hashlib
from math import ceil, exp, log

from bitarray import bitarray

if sys.version_info[0] == 2:
    int = long
    range = xrange
    log2 = lambda x: log(x, 2.0)
else:
    from math import log2


class BloomFilter(object):
    """
    Implementation of a Bloom filter.  An instance is initialized by
    it's capacity `n` and error rate `p`.  The capacity tells how many
    elements can be stored while maintaining no more than `p` false
    positives.
    """
    def __init__(self, n, p=0.01):
        assert 0 < p < 1
        self.n = n
        # number of hash functions
        self.k = int(ceil(-log2(p)))
        # size of array
        self.m = int(ceil(-n * log2(p) / log(2)))
        self.array = bitarray(self.m)
        self.array.setall(0)

    def calculate_p(self):
        """
        Calculate the actual false positive error rate `p` from the number
        of hashes `k` and the size if the bitarray `m`.  This is slightly
        different from the given `p`, because the integer value of `k`
        is being used.
        """
        return pow(1.0 - exp(-float(self.k) * self.n / self.m), self.k)

    def approx_items(self):
        """
        Return the approximate number of items in the Bloom filter.
        """
        count = self.array.count()
        if count == 0:
            return 0.0
        return -float(self.m) / self.k * log(1.0 - float(count) / self.m)

    def add(self, key):
        for i in self._hashes(key):
            self.array[i] = 1

    def __contains__(self, key):
        return all(self.array[i] for i in self._hashes(key))

    def _hashes(self, key):
        """
        generate k different hashes, each of which maps a key to one of
        the m array positions with a uniform random distribution
        """
        h = hashlib.new('md5')
        h.update(str(key).encode())
        x = int(h.hexdigest(), 16)
        for _unused in range(self.k):
            if x < 1024 * self.m:
                h.update(b'x')
                x = int(h.hexdigest(), 16)
            x, y = divmod(x, self.m)
            yield y


def test_bloom(n, p):
    print("Testing Bloom filter:")
    print("capacity     n = %d" % n)
    print("given        p = %.3f%%" % (100.0 * p))
    b = BloomFilter(n, p)
    print("hashes       k = %d = ceil(%.3f)" % (b.k, -log2(p)))
    print("array size   m = %d" % b.m)
    for i in range(n):
        b.add(i)
        assert i in b
    print("approx_items(): %.2f" % b.approx_items())
    print("calculate_p(): %.3f%%" % (100.0 * b.calculate_p()))

    N = 100000
    false_pos = sum(i in b for i in range(n, n + N))
    print("experimental : %.3f%%\n" % (100.0 * false_pos / N))


if __name__ == '__main__':
    test_bloom(5000, 0.05)
    test_bloom(10000, 0.01)
    test_bloom(50000, 0.005)
    test_bloom(100000, 0.002)



"""
Implementation of a sparse bitarray

For example:

   bitarray('110011111000')

is represented as:

   length:  11
   ones:    [0, 1, 4, 5, 6, 7, 8]
"""
from bisect import bisect_left

from bitarray import bitarray

from common import Common


class SparseBitarray(Common):

    def __init__(self, x = 0):
        if isinstance(x, int):
            self.n = x
            self.ones = []
        else:
            self.n = 0
            self.ones = []
            for v in x:
                self.append(int(v))

    def __len__(self):
        return self.n

    def __getitem__(self, key):
        if isinstance(key, slice):
            start, stop = self._get_start_stop(key)
            if stop <= start:
                return SparseBitarray()

            i = bisect_left(self.ones, start)
            j = bisect_left(self.ones, stop)
            res = SparseBitarray(stop - start)
            for k in range(i, j):
                res.ones.append(self.ones[k] - start)
            return res

        elif isinstance(key, int):
            if not 0 <= key < self.n:
                raise IndexError
            i = bisect_left(self.ones, key)
            return int(i != len(self.ones) and self.ones[i] == key)

        else:
            raise TypeError

    def __setitem__(self, key, value):
        if isinstance(key, slice):
            start, stop = self._get_start_stop(key)
            if stop <= start:
                return

            i = bisect_left(self.ones, start)
            j = bisect_left(self.ones, stop)
            del self.ones[i:j]
            if value == 0:
                return
            self.ones.extend((range(start, stop)))
            self.ones.sort()

        elif isinstance(key, int):
            if not 0 <= key < self.n:
                raise IndexError
            i = bisect_left(self.ones, key)
            if i != len(self.ones) and self.ones[i] == key:  # key present
                if value == 0:
                    del self.ones[i]
            else:  # key not present
                if value == 1:
                    self.ones.insert(i, key)

        else:
            raise TypeError

    def __delitem__(self, key):
        if isinstance(key, slice):
            start, stop = self._get_start_stop(key)
            if stop <= start:
                return

            i = bisect_left(self.ones, start)
            j = bisect_left(self.ones, stop)
            del self.ones[i:j]
            size = stop - start
            for k in range(i, len(self.ones)):
                self.ones[k] -= size
            self.n -= size

        elif isinstance(key, int):
            if not 0 <= key < len(self):
                raise IndexError
            i = bisect_left(self.ones, key)
            if i != len(self.ones) and self.ones[i] == key:
                del self.ones[i]
            for k in range(i, len(self.ones)):
                self.ones[k] -= 1
            self.n -= 1

        else:
            raise TypeError

    def append(self, value):
        if value:
            self.ones.append(self.n)
        self.n += 1

    def find(self, value):
        ones = self.ones
        if value:
            return ones[0] if ones else -1
        else:
            m = len(ones)
            if m == self.n:
                return -1
            for i in range(m):
                if ones[i] != i:
                    return i
            return m

    def extend(self, other):
        self.ones.extend(other.ones[i] + self.n for i in
                         range(len(other.ones)))
        self.n += other.n

    def to_bitarray(self):
        a = bitarray(self.n)
        a.setall(0)
        for k in self.ones:
            a[k] = 1
        return a

    def insert(self, k, value):
        k = self._adjust_index(k)
        i = bisect_left(self.ones, k)
        for j in range(i, len(self.ones)):
            self.ones[j] += 1
        self.n += 1
        self[k] = value

    def invert(self):
        self.ones = sorted(set(range(self.n)) - set(self.ones))

    def count(self, value=1):
        if value:
            return len(self.ones)
        else:
            return self.n - len(self.ones)

    def reverse(self):
        lst = [self.n - i - 1 for i in self.ones]
        lst.reverse()
        self.ones = lst



import sys
import datetime
import zlib
from pprint import pprint

from bitarray import bitarray

from puff import Puff


class GunZip(Puff):

    operating_system = {
        0: "FAT",      1: "Amiga",          2: "VMS",      3: "Unix",
        4: "VM/CMS",   5: "Atari TOS",      6: "HPFS",     7: "Macintosh",
        8: "Z-System", 9: "CP/M",          10: "TOPS-20", 11: "NTFS",
        12: "QDOS",   13: "Acorn RISCOS", 255: "Unknown",
    }

    def read_null_terminated_string(self) -> str:
        a = bytearray()
        while True:
            b: int = self.read_uint(8)
            if b == 0:
                return a.decode("UTF-8")
            a.append(b)

    def read_header(self) -> None:
        if self.read_uint(16) != 0x8b1f:
            raise ValueError("Invalid GZIP magic number")

        cmeth = self.read_uint(8)
        if cmeth != 8:
            raise ValueError(f"Unsupported compression method: {str(cmeth)}")

        # reserved flags
        flags: int = self.read_uint(8)
        if flags & 0xe0 != 0:
            print("Reserved flags are set")

        # modification time
        mtime = self.read_uint(32)
        if mtime != 0:
            dt = datetime.datetime.fromtimestamp(mtime, datetime.timezone.utc)
            print(f"Last modified: {dt}")
        else:
            print("Last modified: N/A")

        # extra flags
        extraflags = self.read_uint(8)
        if extraflags == 2:
            print("Extra flags: Maximum compression")
        elif extraflags == 4:
            print("Extra flags: Fastest compression")
        else:
            print(f"Extra flags: Unknown ({extraflags})")

        osbyte = self.read_uint(8)
        osstr: str = self.operating_system.get(osbyte, "Really unknown")
        print(f"Operating system: {osstr}")

        # handle assorted flags
        if flags & 0x01:
            print("Flag: Text")
        if flags & 0x04:
            print("Flag: Extra")
            count: int = self.read_uint(16)
            while count > 0:  # Skip extra data
                self.read_uint(8)
                count -= 1
        if flags & 0x08:
            print(f"File name: {self.read_null_terminated_string()}")
        if flags & 0x02:
            print(f"Header CRC-16: {self.read_uint(16):04X}")
        if flags & 0x10:
            print(f"Comment: {self.read_null_terminated_string()}")

    def check_footer(self, decomp):
        self.align_byte_boundary()

        crc = self.read_uint(32)
        size = self.read_uint(32)

        # check decompressed data's length and CRC
        if size != len(decomp):
            raise ValueError(f"Size mismatch: expected={size}, "
                             f"actual={len(decomp)}")

        actualcrc = zlib.crc32(decomp) & 0xffffffff
        if crc != actualcrc:
            raise ValueError(f"CRC-32 mismatch: expected={crc:08X}, "
                             f"actual={actualcrc:08X}")

def print_dot(*args):
    sys.stdout.write('.')
    sys.stdout.flush()

def decompress_file(infile, outfile, opts):
    # read input file and store content in little endian bitarray
    input_bits = bitarray(0, 'little')
    with open(infile, "rb") as fi:
        input_bits.fromfile(fi)

    # gunzip: the output is accumulated in a bytearray
    output = bytearray()
    d = GunZip(input_bits, output)
    d.read_header()
    stats = d.process_blocks(print_dot if opts.progress else None)
    d.check_footer(output)

    if opts.progress:
        sys.stdout.write('\n')
    if opts.stats:
        pprint(stats)

    # write output to file
    with open(outfile, "wb") as fo:
        fo.write(output)


def main():
    from optparse import OptionParser

    p = OptionParser(usage="usage: %prog [options] IN_FILE.gz [OUT_FILE]")

    p.add_option('-p', '--progress',
                 action="store_true",
                 help="show progress while decoding")

    p.add_option('-s', '--stats',
                 action="store_true",
                 help="show block statistics")

    opts, args = p.parse_args()

    if len(args) not in (1, 2):
        p.error("incorrect number of arguments")

    out_file = args[1] if len(args) == 2 else 'out'

    decompress_file(args[0], out_file, opts)


if __name__ == "__main__":
    main()



class Common:

    def __repr__(self):
        return "SparseBitarray('%s')" % (''.join(str(v) for v in self))

    def pop(self, i = -1):
        if i < 0:
            i += len(self)
        res = self[i]
        del self[i]
        return res

    def remove(self, value):
        i = self.find(value)
        if i < 0:
            raise ValueError
        del self[i]

    def sort(self, reverse=False):
        if reverse:
            c1 = self.count(1)
            self[:c1:] = 1
            self[c1::] = 0
        else:
            c0 = self.count(0)
            self[:c0:] = 0
            self[c0::] = 1

    def _get_start_stop(self, key):
        if key.step not in (1, None):
            raise ValueError("only step = 1 allowed, got %r" % key)
        start = key.start
        if start is None:
            start = 0
        stop = key.stop
        if stop is None:
            stop = len(self)
        return start, stop

    def _adjust_index(self, i):
        n = len(self)
        if i < 0:
            i += n
            if i < 0:
                i = 0
        elif i > n:
            i = n
        return i



from collections import Counter

from bitarray import bitarray
from bitarray.util import canonical_huffman, canonical_decode

from huffman import write_dot, print_code, make_tree


def main():
    from optparse import OptionParser

    p = OptionParser("usage: %prog [options] [FILE]")
    p.add_option(
        '-p', '--print',
        action="store_true",
        help="print Huffman code")
    p.add_option(
        '-t', '--tree',
        action="store_true",
        help="store the tree as a .dot file")
    opts, args = p.parse_args()

    if len(args) == 0:
        filename = 'README'
    elif len(args) == 1:
        filename = args[0]
    else:
        p.error('only one argument expected')

    with open(filename, 'rb') as fi:
        plain = bytearray(fi.read())

    freq = Counter(plain)
    code, count, symbol = canonical_huffman(freq)
    if opts.print:
        print_code(freq, code)
    if opts.tree:
        # create tree from code (no frequencies)
        write_dot(make_tree(code), 'tree_raw.dot', 0 in plain)

    a = bitarray()
    a.encode(code, plain)
    assert bytearray(a.iterdecode(code)) == plain
    assert bytearray(canonical_decode(a, count, symbol)) == plain


if __name__ == '__main__':
    main()



"""
see: https://www.youtube.com/watch?v=umTbivyJoiI
"""
import sys
from math import log2
from collections import Counter

from bitarray.util import huffman_code


def efficiency(freq, code):
    total = sum(freq.values())  # total frequency
    H = 0.0                     # entropy
    L = 0.0                     # average length
    for s in freq.keys():
        p = freq[s] / total     # probability
        H -= p * log2(p)
        L += p * len(code[s])
    print('H =', H)
    print('L =', L)
    return H / L                # efficiency

if len(sys.argv) > 1:
    with open(sys.argv[1], 'rb') as fi:
        plain = fi.read()
else:
    plain = b'aaabbbcde'

freq = Counter(plain)
code = huffman_code(freq)
print("Efficiency =", efficiency(freq, code))



from __future__ import print_function
from bitarray import bitarray
from bitarray.util import ba2int, pprint


# See: https://en.wikipedia.org/wiki/UTF-8

def code_point(u):
    print('character:', u)
    b = u.encode('utf-8')
    print('hexadecimal:', ' '.join('%02x' % i for i in bytearray(b)))
    a = bitarray(endian='big')
    a.frombytes(b)
    pprint(a)

    # calculate binary code point from binary UTF-8 representation
    if a[0:1] == bitarray('0'):
        c = a[1:8]
        assert len(a) == 8
    elif a[0:3] == bitarray('110'):
        c = a[3:8] + a[10:16]
        assert a[8:10] == bitarray('10')
        assert len(a) == 16
    elif a[0:4] == bitarray('1110'):
        c = a[4:8] + a[10:16] + a[18:24]
        assert a[8:10] == a[16:18] == bitarray('10')
        assert len(a) == 24
    elif a[0:5] == bitarray('11110'):
        c = a[5:8] + a[10:16] + a[18:24] + a[26:32]
        assert a[8:10] == a[16:18] == a[24:26] == bitarray('10')
        assert len(a) == 32
    else:
        raise
    code_point = ba2int(c)

    print('code point:', hex(code_point))
    print()


for u in u'\u0024 \u00a2 \u20ac \ud55c \U00010348 \U0010ffff'.split():
    code_point(u)



# Useful functions related to sparse bitarray compression.
# In particular the function sc_stats() which returns the
# frequency of each block type.

import bz2
import gzip
import sys
from time import time
from collections import Counter
from itertools import islice
from random import random, randint

from bitarray import bitarray
from bitarray.util import (
    zeros, urandom,
    serialize, deserialize,
    sc_encode, sc_decode,
    vl_encode, vl_decode,
)

def read_n(n, stream):
    i = 0
    for j in range(n):
        i |= next(stream) << 8 * j
    if i < 0:
        raise ValueError("read %d bytes got negative value: %d" % (n, i))
    return i

def sc_decode_header(stream):
    head = next(stream)
    if head & 0xe0:
        raise ValueError("invalid header: 0x%02x" % head)
    endian = 'big' if head & 0x10 else 'little'
    length = head & 0x0f
    nbits = read_n(length, stream)
    return endian, nbits

def sc_decode_block(stream, stats):
    head = next(stream)
    if head == 0:  # stop byte
        return False

    if head <= 0x80:
        n = 0
        k = head
    elif 0xa0 <= head < 0xc0:
        n = 1
        k = head - 0xa0
    elif 0xc2 <= head <= 0xc4:
        n = head - 0xc0
        k = next(stream)
    else:
        raise ValueError("Invalid block head: 0x%02x" % head)

    stats['blocks'][n] += 1

    # consume block data
    size = max(1, n) * k        # size of block data
    next(islice(stream, size, size), None)

    return True

def sc_stats(stream):
    """sc_stats(stream) -> dict

Decode a compressed byte stream (generated by `sc_encode()` and return
useful statistics.  In particular, the frequency of each block type.
"""
    stream = iter(stream)
    endian, nbits = sc_decode_header(stream)

    stats = {
        'endian': endian,
        'nbits': nbits,
        'blocks': Counter()
    }

    while sc_decode_block(stream, stats):
        pass

    stop = False
    try:
        next(stream)
    except StopIteration:
        stop = True
    assert stop

    return stats

def test_sc_stat():
    a = bitarray(1<<33, 'little')
    a.setall(0)
    a[:1<<16] = 1
    a[:1<<18:1<<4] = 1
    a[:1<<22:1<<12] = 1
    a[:1<<30:1<<20] = 1
    assert a.count() == 79804
    b = sc_encode(a)
    stat = sc_stats(b)
    assert stat['endian'] == 'little'
    assert stat['nbits'] == 1 << 33
    blocks = stat['blocks']
    for i, n in enumerate([64, 754, 46, 48, 2]):
        print("         block type %d  %8d" % (i, blocks[i]))
        assert blocks[i] == n
    if sys.version_info[:2] >= (3, 10):
        print("total number of blocks %8d" % blocks.total())
    assert a == sc_decode(b)

def test_raw_block_size():
    for n in range(10_000):
        a = bitarray(n)
        a.setall(1)
        b = sc_encode(a)
        stat = sc_stats(b)
        assert stat['nbits'] == n
        blocks = stat['blocks']
        assert blocks[0] == (n + 1023) // 1024
        assert sc_decode(b) == a

def random_array(n, p=0.5):
    """random_array(n, p=0.5) -> bitarray

Generate random bitarray of length n.
Each bit has a probability p of being 1.
"""
    if p < 0.05:  # XXX what happens for small N?  N=0 crashes right now.
        # when the probability p is small, it is faster to randomly
        # set p * n elements
        a = zeros(n)
        for _ in range(int(p * n)):
            a[randint(0, n - 1)] = 1
        return a

    return bitarray((random() < p for _ in range(n)))

def test_random_array():
    n = 10_000_000
    p = 1e-6
    while p < 1.0:
        a = random_array(n, p)
        cnt = a.count()
        print("%10.7f  %10.7f  %10.7f" % (p, cnt / n, abs(p - cnt / n)))
        p *= 1.4

def p_range():
    n = 1 << 28
    p = 1e-8
    print("        p          ratio         raw"
          "    type 1    type 2    type 3    type 4")
    print("   " + 73 *'-')
    while p < 1.0:
        a = random_array(n, p)
        b = sc_encode(a)
        blocks = sc_stats(b)['blocks']
        print('  %11.8f  %11.8f  %8d  %8d  %8d  %8d  %8d' % (
            p, len(b) / (n / 8),
            blocks[0], blocks[1], blocks[2], blocks[3], blocks[4]))
        assert a == sc_decode(b)
        p *= 1.8

def compare():
    n = 1 << 26
    # create random bitarray with p = 1 / 2^9 = 1 / 512 = 0.195 %
    a = bitarray(n)
    a.setall(1)
    for i in range(10):
        a &= urandom(n)

    raw = a.tobytes()
    print(20 * ' ' +  "compress (ms)   decompress (ms)             ratio")
    print(70 * '-')
    for name, f_e, f_d in [
            ('serialize', serialize, deserialize),
            ('vl', vl_encode, vl_decode),
            ('sc' , sc_encode, sc_decode),
            ('gzip', gzip.compress, gzip.decompress),
            ('bz2', bz2.compress, bz2.decompress)]:
        x = a if name in ('serialize', 'vl', 'sc') else raw
        t0 = time()
        b = f_e(x)  # compression
        t1 = time()
        c = f_d(b)  # decompression
        t2 = time()
        print("    %-11s  %16.3f  %16.3f  %16.4f" %
              (name, 1000 * (t1 - t0), 1000 * (t2 - t1), len(b) / len(raw)))
        assert c == x

if __name__ == '__main__':
    test_sc_stat()
    test_raw_block_size()
    #test_random_array()
    compare()
    p_range()



"""
In this module, we implement distance functions and compare them to the
corresponding functions in the scipy.spatial.distance module.
The functions in this module are typically around 10 to 50 times faster.
"""
from time import time

from bitarray.util import _correspond_all, count_and, count_xor, urandom

import numpy
import scipy.spatial.distance as distance  # type: ignore


def dice(u, v):
    x = float(count_xor(u, v))
    return x / (2.0 * count_and(u, v) + x)

def hamming(u, v):
    return float(count_xor(u, v)) / len(u)

def jaccard(u, v):
    x = float(count_xor(u, v))
    return x / (count_and(u, v) + x)

def kulczynski1(u, v):
    return float(count_and(u, v)) / count_xor(u, v)

def rogerstanimoto(u, v):
    x = float(count_xor(u, v))
    return 2.0 * x / (len(u) + x)

def russellrao(u, v):
    n = float(len(u))
    return (n - count_and(u, v)) / n

def sokalmichener(u, v):
    x = float(count_xor(u, v))
    return 2.0 * x / (len(u) + x)

def sokalsneath(u, v):
    R = 2.0 * count_xor(u, v)
    return R / (count_and(u, v) + R)

def yule(u, v):
    nff, nft, ntf, ntt = _correspond_all(u, v)
    half_R = ntf * nft
    if half_R == 0:
        return 0.0
    else:
        return 2.0 * half_R / (ntt * nff + half_R)


def test(n):
    a = urandom(n)
    b = urandom(n)
    aa = numpy.frombuffer(a.unpack(), dtype=bool)
    bb = numpy.frombuffer(b.unpack(), dtype=bool)

    for name in ['dice', 'hamming', 'jaccard', 'kulczynski1',
                 'rogerstanimoto', 'russellrao', 'sokalmichener',
                 'sokalsneath', 'yule']:
        f1 = eval(name)
        t0 = time()
        x1 = f1(a, b)
        t1 = time() - t0
        print('%.14f  %9.6f sec  %s' % (x1, t1, name))

        f2 = getattr(distance, name)
        t0 = time()
        x2 = f2(aa, bb)
        t2 = time() - t0
        print('%.14f  %9.6f sec  %9.2f' % (x2, t2, t2 / t1))

        assert abs(x1 - x2) < 1E-14

test(2 ** 25 + 67)
